"""genestboost.utils module."""

# author: Benjamin Cross
# email: btcross26@yahoo.com
# created: 2019-12-16

from .gradients import d1_central_difference, d2_central_difference

__all__ = ["d1_central_difference", "d2_central_difference"]
